/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : BooG - tree permutations
 * Date    : 2000
 * Author  : Francois Donnet
 */


#include MUT_H
#include ABL_H
#include ABE_H
#include "bog_normalize_ARITY.h"
#include "bog_lib_matching.h"
#include "bog_lib_specifications.h"
#include "bog_lib_utils.h"
#include "bog_lib_permute.h"



/******************************************************************************/
/* duplicate a list of abl                                                    */
/******************************************************************************/
static chain_list* dupablchain(chain_list* ablchain)
{
   chain_list* chain,*res=NULL;

   for (chain=ablchain;chain;chain=chain->NEXT) {
      res=addchain(res,dupablexpr(chain->DATA));
   }
   
   return res;
}


/******************************************************************************/
/*return a list of abl with abl element put in all positions of ablchain      */
/******************************************************************************/
static chain_list *all_positions(chain_list* head, chain_list* ablchain, chain_list* abl)
{
  chain_list *chain, *list_abl, *new, *pred;
  
  if (!ablchain) {
      ablchain=addchain(NULL,dupablexpr(abl));
      return addchain(head,ablchain);
   }   
      
   /*insert abl on a lower position*/
   list_abl=all_positions(NULL,ablchain->NEXT,abl);

   /*add the first element of ablchain for the result*/
   for (chain=list_abl; chain; chain=chain->NEXT) {
      new=dupablexpr(ABL_CAR(ablchain));
      chain->DATA=addchain(chain->DATA,new);
      pred=chain;
   }

   /*create a new element with abl placed in first*/
   ablchain=addchain(dupablchain(ablchain),dupablexpr(abl));
   list_abl=addchain(list_abl,ablchain);
   
   /*take in account the head*/
   pred->NEXT=head;
   
   return list_abl;
}


/******************************************************************************/
/*permute all elements of ablchain and return the list of permutations        */
/******************************************************************************/
static chain_list *swap(chain_list* head, chain_list* ablchain)
{
  chain_list *chain, *list_abl, *swap_abl;
  chain_list *first;
  
   
  if (!ablchain) {
      if (!head) return addchain(NULL,NULL);
      return head;
   }   

   first=ABL_CAR(ablchain);
   /*return a list of all possibilities with the rest of ablchain*/
   swap_abl=swap(NULL,ablchain->NEXT);

   list_abl=head;
   for (chain=swap_abl; chain; chain=chain->NEXT) {
      ablchain=chain->DATA;
      /*insert new element in all positions*/
      /*add it to final result*/
      list_abl=all_positions(list_abl,ablchain,first);
   }
   
   /*free mem*/
   for (chain=swap_abl; chain; chain=chain->NEXT) freeablexpr(chain->DATA);
   freechain(swap_abl);

   return list_abl;
}



/******************************************************************************/
/*return the list of permutations based on the leaf choice (parameter)        */
/* leaves_list is a list of permutations of each leaf                         */
/******************************************************************************/
static chain_list* compose(chain_list* leaves_list)
{
   chain_list *leaf, *list_abl;
   chain_list *abl;
   chain_list* new_list;
   chain_list *chain, *ablchain; 

   /*the list is empty*/
   if (!leaves_list) return addchain(NULL,NULL);

   list_abl=compose(leaves_list->NEXT);

   new_list=NULL;
   /*add current leaf*/
   for (leaf=leaves_list->DATA; leaf; leaf=leaf->NEXT) {

      /*compose with the list of different abl of the rest of leaves*/
      for (chain=list_abl; chain; chain=chain->NEXT) {
         ablchain=chain->DATA;
         abl=leaf->DATA;
         if (ablchain) ablchain=dupablchain(ablchain); /*only if not NULL*/
         abl=dupablexpr(abl);
         ablchain=addchain(ablchain,abl);
         new_list=addchain(new_list,ablchain);
      }
      
   }

   /*free mem*/
   for (chain=list_abl; chain; chain=chain->NEXT) freeablexpr(chain->DATA);
   freechain(list_abl);
   
   return new_list;
}


/******************************************************************************/
/*return a list of abl. It is the list of all possible permutations of an abl */
/*abl parameter isn't deleted                                                 */
/******************************************************************************/
static chain_list *permutations(chain_list* abl)
{
   chain_list *head, *list_abl, *permutations_leaves_list;
   chain_list *leaf, *chain, *compose_abl, *ablchain;
   
   /*leaf*/
   if (ABL_ATOM(abl)) return addchain(NULL,dupablexpr(abl));

   /*build first permutations on leaves and get back*/
   permutations_leaves_list=NULL;
   for (chain=ABL_CDR(abl); chain; chain=ABL_CDR(chain)) {
      /*list of list of permutated leaves*/
      permutations_leaves_list=addchain(permutations_leaves_list,
                                        permutations(ABL_CAR(chain)));
   }
   
   /*build all the possible abl with this choice of leaves*/
   compose_abl=compose(permutations_leaves_list);
   list_abl=NULL;
   /*build all permutations with this composition*/
   for (chain=compose_abl; chain; chain=chain->NEXT) {
      ablchain=chain->DATA;
      /*list of permutations with these elements of ablchain*/
      /*add it to final result*/
      list_abl=swap(list_abl,ablchain);
   }
   
   /*put the operator*/
   for (chain=list_abl; chain; chain=chain->NEXT) {
      head=createabloper(ABL_OPER(abl));
      ABL_ARITY(head)=ABL_ARITY(abl);
      ABL_CDR(head)=chain->DATA;
      chain->DATA=head;
   }   

   /*free previous search*/
   for (chain=permutations_leaves_list; chain; chain=chain->NEXT) {
      /*contains a list of abl*/
      for (leaf=chain->DATA; leaf; leaf=leaf->NEXT) {
      freeablexpr(leaf->DATA);
      }
      freechain(chain->DATA);
   }
   for (chain=compose_abl; chain; chain=chain->NEXT) {
      /*contains abl*/
      freeablexpr(chain->DATA);
   }
   freechain(permutations_leaves_list);
   freechain(compose_abl);
   
   return list_abl;
}


/******************************************************************************/
/* return 1 if abl1 and abl2 heave the same properties                        */
/******************************************************************************/
static int same_properties(chain_list* abl1, chain_list* abl2, befig_list* befig)
{
   if (ABL_ATOM(abl1)!=ABL_ATOM(abl2)) {
      fprintf(stderr,"same_properties: compute error\n");
      exit(1);
   }

   if (ABL_ATOM(abl1)) {
      /*no use to compare if the same leaf*/
      if (ABL_ATOM_VALUE(abl1)==ABL_ATOM_VALUE(abl2)) return 0;
      
      /*comparison*/
      if (getgenericT(befig,ABL_ATOM_VALUE(abl1))
         !=getgenericT(befig,ABL_ATOM_VALUE(abl2))) return 0;
      
      if (getgenericR(befig,ABL_ATOM_VALUE(abl1))   /* Rup+Rdown/2 */
         !=getgenericR(befig,ABL_ATOM_VALUE(abl2))) return 0;
      
      if (getgenericC(befig,ABL_ATOM_VALUE(abl1))   /* entry Capacitance */
         !=getgenericC(befig,ABL_ATOM_VALUE(abl2))) return 0;

      return 1;   /*equality*/
   }

   /*operator*/
   abl2=ABL_CDR(abl2);
   for (abl1=ABL_CDR(abl1); abl1; abl1=ABL_CDR(abl1)) {
      if (!same_properties(ABL_CAR(abl1),ABL_CAR(abl2),befig)) return 0;
      abl2=ABL_CDR(abl2);
   }
   
   return 0;   /*equality*/
}


/******************************************************************************/
/*    return the list of different abl   resulting of permutations            */
/******************************************************************************/
extern chain_list* different_abl(befig_list* befig, chain_list* abl)
{
   chain_list* list_abl;

   /*get all permutations possible*/
   list_abl=permutations(abl);

#if 0   /*no use, all abl are different*/
   bepor_list *bepor, *bepor_aux;
   chain_list* chain;
   chain_list* control, *pred;
   
   for (bepor=befig->BEPOR; bepor; bepor=bepor->NEXT) {
      if (isvdd(bepor->NAME) || isvss(bepor->NAME)) continue;
      if (bepor->DIRECTION=='O' || bepor->DIRECTION=='Z') continue;
      for (bepor_aux=befig->BEPOR; bepor_aux; bepor_aux=bepor_aux->NEXT) {
        if (isvdd(bepor_aux->NAME) || isvss(bepor_aux->NAME)) continue;
        if (bepor_aux->DIRECTION=='O' || bepor_aux->DIRECTION=='Z') continue;

        /*search a leaves with equal properties*/
        if (getgenericT(befig,bepor->NAME)==getgenericT(befig,bepor_aux->NAME)
         && getgenericR(befig,bepor->NAME)==getgenericR(befig,bepor_aux->NAME)
         && getgenericC(befig,bepor->NAME)==getgenericC(befig,bepor_aux->NAME)){
           
            /*put off duplicate use*/
            for (control=list_abl; control; control=control->NEXT) {
               pred=control;
               for (chain=control->NEXT; chain; chain=chain->NEXT) {
                  /*erase doublon*/
                  if (same_properties(control->DATA,chain->DATA,befig)) {
                     freeablexpr(chain->DATA);
                     pred->NEXT=chain->NEXT;
                     chain->NEXT=NULL;
                     freechain(chain);
                     chain=pred;
                  }   
                  pred=chain;
               }
            } /*end of loop*/
        
        }
     }/*end of loop*/
   }/*end of loop*/
#endif   
   
   return list_abl;
}


